package metrik.metrics.domain.calculator

import com.fasterxml.jackson.module.kotlin.jacksonObjectMapper
import com.fasterxml.jackson.module.kotlin.readValue
import metrik.metrics.domain.model.LEVEL
import metrik.project.domain.model.Execution
import org.assertj.core.api.Assertions.assertThat
import org.junit.jupiter.api.Assertions.assertEquals
import org.junit.jupiter.api.BeforeEach
import org.junit.jupiter.api.Test

internal class DeploymentFrequencyCalculatorTest {
    private lateinit var deploymentFrequencyCalculator: DeploymentFrequencyCalculator

    @BeforeEach
    fun setUp() {
        deploymentFrequencyCalculator = DeploymentFrequencyCalculator()
    }

    /**
     * test file: builds-for-DF-case-1.json
     * pipeline 1
     * build 1 : deploy to prod, SUCCESS, start at 2021-01-01, end at 2021-01-01
     * build 2 : deploy to prod, SUCCESS, start at 2021-01-15, end at 2021-01-15
     * build 3 : deploy to prod, SUCCESS, start at 2021-01-30, end at 2021-02-01
     * pipeline 2
     * build 1 : deploy to prod, SUCCESS, start at 2021-01-01, end at 2021-01-01
     * build 2 : deploy to prod, SUCCESS, start at 2021-01-15, end at 2021-01-15
     * build 3 : deploy to prod, SUCCESS, start at 2021-01-30, end at 2021-02-01
     */
    @Test
    fun `should return DF given all builds with some Aborted and others success`() {
        val allExecutions: List<Execution> = jacksonObjectMapper().readValue(
            this.javaClass.getResource("/calculator/builds-for-DF-case-1.json").readText()
        )

        val startTimestamp = 1610236800000L // 2021-01-10
        val endTimestamp = 1611100800000L // 2021-01-20
        val targetStage = mapOf(Pair("1", "deploy to prod"), Pair("2", "deploy to prod"))

        val deploymentCount =
            deploymentFrequencyCalculator.calculateValue(allExecutions, startTimestamp, endTimestamp, targetStage)

        assertEquals(deploymentCount, 2)
    }

    /**
     * test file: builds-for-DF-case-3.json
     * build 1 : 2021-01-01, [build: SUCCESS, deploy to prod: SUCCESS]
     * build 2 : 2021-01-15, [build: SUCCESS, deploy to prod: FAILED]
     * build 3 : 2021-01-30, [build: FAILED, deploy to prod: SUCCESS]
     */
    @Test
    fun `should get deployment count with success target stage status when calculate deployment count`() {
        val targetStage = mapOf(Pair("1", "deploy to prod"), Pair("2", "deploy to prod"))
        val startTimestamp = 1609286400000L // 2020-12-30
        val endTimestamp = 1612137600000L // 2021-02-01

        val mockExecutionList: List<Execution> =
            jacksonObjectMapper().readValue(
                this.javaClass.getResource("/calculator/builds-for-DF-case-3.json").readText()
            )

        assertThat(
            deploymentFrequencyCalculator.calculateValue(
                mockExecutionList,
                startTimestamp,
                endTimestamp,
                targetStage
            )
        ).isEqualTo(2)
    }

    @Test
    fun `should return elite given input deployment count greater than 15 when calculated in fortnightly`() {
        val deployments = 15
        val days = 14

        assertThat(deploymentFrequencyCalculator.calculateLevel(deployments, days)).isEqualTo(LEVEL.ELITE)
    }

    @Test
    fun `should return high given input deployment frequency within 1 div 7 and 1`() {
        val deployments = 1
        val days = 3

        assertThat(deploymentFrequencyCalculator.calculateLevel(deployments, days)).isEqualTo(LEVEL.HIGH)
    }

    @Test
    fun `should return medium given input deployment frequency within 1 div 30 and 1 div 7`() {
        val deployments = 1
        val days = 7

        assertThat(deploymentFrequencyCalculator.calculateLevel(deployments, days)).isEqualTo(LEVEL.MEDIUM)
    }

    @Test
    fun `should return low given input deployment frequency greater than`() {
        val deployments = 1
        val days = 30

        assertThat(deploymentFrequencyCalculator.calculateLevel(deployments, days)).isEqualTo(LEVEL.LOW)
    }
}
